/* Parse an hourly connection log file for Zoom info */
/* and summarize each of the Zoom sessions found.    */
/* Builds a large and sparse traffic matrix to keep  */
/* track of different clients and Zoom servers.      */
/* Beware of memory usage issues on big traces!      */

/* Written by Carey Williamson          October 2021 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Manifest constants */
#define MAX_CLIENTS 300   /* be careful if increasing this! */
#define MAX_SERVERS 100   /* be careful if increasing this! */
#define IP_LEN 16
#define LEN 8
#define CLIENT_PREFIX "136.159"   /* UCalgary's /16 prefix */
#define PREFIX_LENGTH 7

/* Choose one of these for either seconds or minutes */
#define TIME_SCALING 1.0   /* show everything in seconds */
//#define TIME_SCALING 60.0  /* convert seconds to minutes in output */

/* Thresholds for UDP channel classification */
#define DATA_THRESHOLD 75
#define AUDIO_THRESHOLD 80
#define VIDEO_THRESHOLD 500
#define PKT_RATE_THRESHOLD 2
#define TOO_SHORT 1.0  /* seconds */

/* Flags to help count channel classifications within a session */
#define UNKNOWN 0x1000
#define AUDIO 0x0100
#define VIDEO 0x0010
#define DATA 0x0001

/* Report connection details for post-processing or graphing */
#define VERBOSE 1

/* Verbose debugging when needed */
/* #define DEBUG 1 */

int main()
  {
    double newtime, oldtime, starttime;
    char leftip[IP_LEN], rightip[IP_LEN], tempip[IP_LEN], lastip[IP_LEN];
    char clientiplist[MAX_CLIENTS][IP_LEN], serveriplist[MAX_SERVERS][IP_LEN];
    int leftport, rightport, tempport, diff;
    int psent, precd, bsent, brecd, temp;
    float duration;
    char protocol[LEN], state[LEN];
    int line, clientcount, servercount, swapped, i;
    float avgsent, avgrecd, sentrate, recdrate, index;
    int x, y, zooms;
    unsigned short int tcpconns[MAX_CLIENTS][MAX_SERVERS], udpconns[MAX_CLIENTS][MAX_SERVERS], icmpconns[MAX_CLIENTS][MAX_SERVERS];
    float tcpstart[MAX_CLIENTS][MAX_SERVERS], udpstart[MAX_CLIENTS][MAX_SERVERS], icmpstart[MAX_CLIENTS][MAX_SERVERS];
    float tcpend[MAX_CLIENTS][MAX_SERVERS], udpend[MAX_CLIENTS][MAX_SERVERS], icmpend[MAX_CLIENTS][MAX_SERVERS];
    unsigned short int tcpbelow1[MAX_CLIENTS][MAX_SERVERS], tcpbelow10[MAX_CLIENTS][MAX_SERVERS], tcpbelow100[MAX_CLIENTS][MAX_SERVERS],
      tcpbelow1000[MAX_CLIENTS][MAX_SERVERS], tcpover1000[MAX_CLIENTS][MAX_SERVERS];
    unsigned short int udpbelow1[MAX_CLIENTS][MAX_SERVERS], udpbelow10[MAX_CLIENTS][MAX_SERVERS], udpbelow100[MAX_CLIENTS][MAX_SERVERS],
      udpbelow1000[MAX_CLIENTS][MAX_SERVERS], udpover1000[MAX_CLIENTS][MAX_SERVERS];
    unsigned short int udpport[MAX_CLIENTS][MAX_SERVERS];
    unsigned short int windows[MAX_CLIENTS][MAX_SERVERS];
    unsigned short int session[MAX_CLIENTS][MAX_SERVERS];
    int totalKBsent, totalKBrecd;
    float firstone, lastone;
    
    /* initialization */
    bzero(leftip, IP_LEN);
    bzero(rightip, IP_LEN);
    bzero(tempip, IP_LEN);
    bzero(lastip, IP_LEN);
    bzero(protocol, LEN);
    bzero(state, LEN);
    line = 0;
    clientcount = 0;
    servercount = 0;
    zooms = 0;
    totalKBsent = 0;
    totalKBrecd = 0;
    firstone = 0.0;
    lastone = 0.0;
    oldtime = 0.0;
    for( x = 0; x < MAX_CLIENTS; x++ )
      {
	for( y = 0; y < MAX_SERVERS; y++ )
	  {
	    tcpconns[x][y] = 0;
	    udpconns[x][y] = 0;
	    icmpconns[x][y] = 0;
	    tcpstart[x][y] = 0.0;
	    tcpend[x][y] = 0.0;
	    udpstart[x][y] = 0.0;
	    udpend[x][y] = 0.0;
	    icmpstart[x][y] = 0.0;
	    icmpend[x][y] = 0.0;
	    tcpbelow1[x][y] = 0;
	    tcpbelow10[x][y] = 0;
	    tcpbelow100[x][y] = 0;
	    tcpbelow1000[x][y] = 0;
	    tcpover1000[x][y] = 0;
	    udpbelow1[x][y] = 0;
	    udpbelow10[x][y] = 0;
	    udpbelow100[x][y] = 0;
	    udpbelow1000[x][y] = 0;
	    udpover1000[x][y] = 0;
	    udpport[x][y] = 0;
	    windows[x][y] = 0;
	    session[x][y] = 0;
	  }
      }

    /* scan the log file */
    while( scanf("%lf %s %d %s %d %s %f %s %d %d %d %d\n",
		 &newtime, leftip, &leftport, rightip, &rightport, protocol,
		 &duration, state, &psent, &bsent, &precd, &brecd) != EOF )
      {
	line++;
	if( line == 1 )
	  starttime = newtime;

#ifdef DEBUG
	printf("protocol field on line %d for host %s to host %s is initially '%s'\n",
	       line, leftip, rightip, protocol);
#endif

	/* make sure that client IP is always on the left (originator) */
	swapped = 0;
	if( (strcmp(leftip, CLIENT_PREFIX) < PREFIX_LENGTH) )
	  {
	    /* switch left IP and right IP */
	    strcpy(tempip, leftip);
	    strcpy(leftip, rightip);
	    strcpy(rightip, tempip);
	    /* port numbers */
	    temp = leftport;
	    leftport = rightport;
	    rightport = temp;
	    /* sent and received values for packets and for bytes */
	    temp = psent;
	    psent = precd;
	    precd = temp;
	    temp = bsent;
	    bsent = brecd;
	    brecd = temp;
	    swapped = 1;
	  }

	/* update list of distinct client IPs seen, if needed */
	index = 0.0;
	for( i = 0; i < clientcount; i++ )
	  {
	    if( strcmp(leftip, clientiplist[i]) == 0 )
	      {
#ifdef DEBUG
		printf("IP %s is already on clientiplist in position %d\n", leftip, i);
#endif
		x = i;
		index = i+1.0;
		break;
	      }
	  }
	if( i == clientcount )
	  {
#ifdef DEBUG
	    printf("adding IP %s to clientiplist in position %d\n", leftip, i);
#endif
	    strcpy(clientiplist[i], leftip);
	    x = i;
	    clientcount++;
	    index = clientcount+0.0;
	    if( clientcount == MAX_CLIENTS )
	      {
		printf("Warning! MAX_CLIENTS %d reached on line %d\n", MAX_CLIENTS, line);
		exit(1);
	      }
	  }
	/* update list of distinct server IPs seen, if needed */
	for( i = 0; i < servercount; i++ )
	  {
	    if( strcmp(rightip, serveriplist[i]) == 0 )
	      {
#ifdef DEBUG
		printf("IP %s is already on serveriplist in position %d\n", rightip, i);
#endif
		y = i;
		break;
	      }
	  }
	if( i == servercount )
	  {
#ifdef DEBUG
	    printf("adding IP %s to serveriplist in position %d\n", rightip, i);
#endif
	    strcpy(serveriplist[i], rightip);
	    y = i;
	    servercount++;
	    if( servercount == MAX_SERVERS )
	      {
		printf("Warning! MAX_SERVERS %d reached on line %d\n", MAX_SERVERS, line);
		exit(1);
	      }
	  }
#ifdef DEBUG
	printf("protocol field on line %d for host %s to host %s is still '%s'\n",
	       line, leftip, rightip, protocol);
#endif
	/* update cumulative bytes sent and received */
	totalKBsent += (bsent>>10);
	totalKBrecd += (brecd>>10);
	
	/* nudge the index to improve graph readability */
	if( strcmp(protocol, "tcp") == 0 )
	  {
	    if( tcpconns[x][y] == 0 )
	      tcpstart[x][y] = newtime - starttime;
	    if( newtime - starttime + duration > tcpend[x][y] )
	      tcpend[x][y] = newtime - starttime + duration;
	    if( tcpend[x][y] > lastone )
	      lastone = tcpend[x][y];
	    tcpconns[x][y]++;
	    index += 0.1;

	    /* record stats regarding relative durations */
	    if( duration < 1.0 )
	      tcpbelow1[x][y]++;
	    else if( duration < 10.0 )
	      tcpbelow10[x][y]++;
	    else if( duration < 100.0 )
	      tcpbelow100[x][y]++;
	    else if( duration < 1000.0 )
	      tcpbelow1000[x][y]++;
	    else tcpover1000[x][y]++;
	  }
	else if( strcmp(protocol, "udp") == 0 )
	  {
	    if( udpconns[x][y] == 0 )
	      udpstart[x][y] = newtime - starttime;
	    if( newtime - starttime + duration > udpend[x][y] )
	      udpend[x][y] = newtime - starttime + duration;
	    if( udpend[x][y] > lastone )
	      lastone = udpend[x][y];
	    udpconns[x][y]++;
	    index -= 0.1;

	    /* record stats regarding relative durations */
	    if( duration < 1.0 )
	      udpbelow1[x][y]++;
	    else if( duration < 10.0 )
	      udpbelow10[x][y]++;
	    else if( duration < 100.0 )
	      udpbelow100[x][y]++;
	    else if( duration < 1000.0 )
	      udpbelow1000[x][y]++;
	    else udpover1000[x][y]++;

	    /* record the client's UDP port, and check if sequential or not */
#ifdef DEBUG
	    printf("Current UDP port: %d   Prev UDP port: %d\n",
		   leftport, udpport[x][y]);
#endif
	    diff = leftport - udpport[x][y];
	    if( diff < 0 )
	      diff *= -1;
	    if( diff <= 2 )
	      windows[x][y]++;
	    udpport[x][y] = leftport;
	  }
	else
	  {
	    if( icmpconns[x][y] == 0 )
	      icmpstart[x][y] = newtime - starttime;
	    if( newtime - starttime + duration > icmpend[x][y] )
	      icmpend[x][y] = newtime - starttime + duration;
	    if( icmpend[x][y] > lastone )
	      lastone = icmpend[x][y];
	    icmpconns[x][y]++;
	    index -= 0.0;
	  }

#ifdef VERBOSE
	/* print out this line of the log */
	printf("%lf %d %3.1f %s %d %s %d %s %f %f %f %s %d %d %d %d",
	       (newtime - starttime)/TIME_SCALING, line, index,
	       leftip, leftport, rightip, rightport, protocol,
	       duration/TIME_SCALING, (newtime - starttime)/TIME_SCALING,
	       (newtime - starttime + duration)/TIME_SCALING,
	       state, psent, bsent, precd, brecd);
#endif

	/* calculate average packet sizes in each direction */
	if( psent > 0 )
	  avgsent = 1.0*bsent/psent;
	else avgsent = 0.0;
	if( precd > 0 )
	  avgrecd = 1.0*brecd/precd;
	else avgrecd = 0.0;

	/* calculate average packet rates in each direction */
	if( duration > TOO_SHORT )
	  sentrate = 1.0*psent/duration;
	else sentrate = 0.0;
	if( duration > TOO_SHORT )
	  recdrate = 1.0*precd/duration;
	else recdrate = 0.0;

#ifdef VERBOSE
	printf(" %3.1f", avgsent);
	printf(" %3.1f", avgrecd);
	printf(" %3.1f", sentrate);
	printf(" %3.1f", recdrate);

	/* use heuristics to tag some of these channels */
	if( avgsent > avgrecd )
	  printf(" P");     /* presenter? */
	else printf(" S");  /* student? */

	if( protocol[0] == 't' )
	  printf("C"); /* TCP control channel */
	else if( protocol[0] == 'i' )
	  printf("I"); /* ICMP */
	/* else it must be UDP */
	else if( duration < TOO_SHORT )
	  printf("X"); /* too short to be sure what it is */
	else if( (avgsent < DATA_THRESHOLD) || (avgrecd < DATA_THRESHOLD) )
	  printf("D"); /* UDP screen sharing data on */
	else if( (avgsent > VIDEO_THRESHOLD) || (avgrecd > VIDEO_THRESHOLD) )
	  printf("V"); /* UDP video */
	else if( (avgsent > AUDIO_THRESHOLD) && (avgrecd > AUDIO_THRESHOLD) )
	  printf("A"); /* UDP audio */
	else if( (sentrate < PKT_RATE_THRESHOLD) && (recdrate < PKT_RATE_THRESHOLD) )
	  printf("D"); /* UDP screen sharing data off */
	else printf("X"); /* not really sure what it is */

	/* tag the lines that got their IPs swapped to help with debugging */
	if( swapped )
	  printf(" *");
	printf("\n");
#endif

	/* use heuristics to tag the UDP channels and report stats */
	/* normally, there are equal numbers of A, V, and D (but not always!) */
	if( protocol[0] == 'u' )
	  {
	    if( duration < TOO_SHORT )
	      session[x][y] += UNKNOWN; /* too short to be sure what it is */
	    if( (avgsent < DATA_THRESHOLD) || (avgrecd < DATA_THRESHOLD) )
	      session[x][y] += DATA;
	    else if( (avgsent > VIDEO_THRESHOLD) || (avgrecd > VIDEO_THRESHOLD) )
	      session[x][y] += VIDEO;
	    else if( (avgsent > AUDIO_THRESHOLD) && (avgrecd > AUDIO_THRESHOLD) )
	      session[x][y] += AUDIO;
	    else if( (sentrate < PKT_RATE_THRESHOLD) && (recdrate < PKT_RATE_THRESHOLD) )
	      session[x][y] += DATA;
	    else session[x][y] += UNKNOWN;
	  }

	/* track time changes */
	oldtime = newtime;

	/* clean up vars for the next one, if any */
	bzero(leftip, IP_LEN);
	bzero(rightip, IP_LEN);
	bzero(tempip, IP_LEN);
	bzero(lastip, IP_LEN);
	bzero(protocol, LEN);
	bzero(state, LEN);
      }
    /* print summary of zoom sessions seen */
    
    printf("Zoom Summary Information\n");
    printf("Client IPs: %d    Server IPs: %d\n", clientcount, servercount);
    printf("Bytes Sent: %d MB    Bytes Received: %d MB   Total Bytes: %d MB\n",
	   totalKBsent>>10, totalKBrecd>>10, (totalKBsent+totalKBrecd)>>10);
    printf("Start Time: %3.1f   End Time:  %3.1f   Avg Mbps:  %4.2f\n",
	   firstone/TIME_SCALING, lastone/TIME_SCALING, 8.0*1024*(totalKBsent+totalKBrecd)/(lastone*1000000.0));
    zooms = 0;
    for( x = 0; x < clientcount; x++ )
      {
	for( y = 0; y < servercount; y++ )
	  {
	    if( (tcpconns[x][y] > 0) || (udpconns[x][y] > 0) || (icmpconns[x][y] > 0) )
	      {
		printf("Zoom %d: %s %s %d %d %d %3.1f %3.1f  %3.1f %3.1f  %3.1f %3.1f  %d %d %d %d %d  %d %d %d %d %d W%1d A%dV%dD%dX%d\n",
		       zooms, clientiplist[x], serveriplist[y],
		       tcpconns[x][y], udpconns[x][y], icmpconns[x][y],
		       tcpstart[x][y]/TIME_SCALING, tcpend[x][y]/TIME_SCALING,
		       udpstart[x][y]/TIME_SCALING, udpend[x][y]/TIME_SCALING,
		       icmpstart[x][y]/TIME_SCALING, icmpend[x][y]/TIME_SCALING,
		       tcpbelow1[x][y], tcpbelow10[x][y], tcpbelow100[x][y],
		       tcpbelow1000[x][y], tcpover1000[x][y],
		       udpbelow1[x][y], udpbelow10[x][y], udpbelow100[x][y],
		       udpbelow1000[x][y], udpover1000[x][y],
		       windows[x][y], (session[x][y]&0x0f00)/AUDIO,
		       (session[x][y]&0x00f0)/VIDEO, (session[x][y]&0x000f)/DATA, (session[x][y]&0xf000)/UNKNOWN);
		zooms++;
	      }
	  }
      }
  }
