/* Parse a connection log file for Zoom call info */
/* and summarize it in a cleaner more logical way */
/* with relative timestamps, visual indication of */
/* timing gaps, and with the client IP always on  */
/* the left, and Zoom server IP on the right.     */

/* Written by Carey Williamson     September 2021 */

#include <stdio.h>
#include <string.h>

/* Manifest constants */
#define IP_LEN 16
#define LEN 6
#define TIME_GAP 2.0

/* This is the /16 network prefix for UCalgary */
#define CLIENT_PREFIX "136.159"
#define PREFIX_LENGTH 7

int main()
  {
    double newtime, oldtime;
    char leftip[IP_LEN], rightip[IP_LEN], tempip[IP_LEN], lastip[IP_LEN];
    int leftport, rightport, tempport;
    int psent, precd, bsent, brecd, temp;
    float duration;
    char protocol[LEN], state[LEN];
    int line, count, swapped;
    float avgsent, avgrecd;
    
    /* initialization */
    bzero(leftip, IP_LEN);
    bzero(rightip, IP_LEN);
    bzero(tempip, IP_LEN);
    bzero(lastip, IP_LEN);
    bzero(protocol, LEN);
    bzero(state, LEN);
    line = 0;
    oldtime = 0.0;

    /* scan the log file */
    while( scanf("%lf %s %d %s %d %s %f %s %d %d %d %d\n",
		 &newtime, leftip, &leftport, rightip, &rightport, protocol,
		 &duration, state, &psent, &bsent, &precd, &brecd) != EOF )
      {
	line++;

	/* Check to see if client IP is on the left (originator) or not. */
	/* Needed for UDP conns since about 10% of them are "backwards". */
	swapped = 0;
	if( strcmp(leftip, CLIENT_PREFIX) < PREFIX_LENGTH )
	  {
	    /* switch left IP and right IP, as well as the relevant fields */
	    strcpy(tempip, leftip);
	    strcpy(leftip, rightip);
	    strcpy(rightip, tempip);
	    /* port numbers */
	    temp = leftport;
	    leftport = rightport;
	    rightport = temp;
	    /* sent and received values for packets and for bytes */
	    temp = psent;
	    psent = precd;
	    precd = temp;
	    temp = bsent;
	    bsent = brecd;
	    brecd = temp;
	    swapped = 1;
	  }

	/* look for timing gaps that are worth showing visually */
	if( (oldtime > 0.0) && (newtime - oldtime > TIME_GAP) )
	  printf("\n");
	
	/* print out this line of the log */
	/* note: absolute timestamp on first line, and relative timestamps after */
	printf("%lf %s %d %s %d %s %f %s %d %d %d %d",
	       newtime - oldtime, leftip, leftport, rightip, rightport, protocol,
	       duration, state, psent, bsent, precd, brecd);

	/* report average packet sizes in each direction (might be useful) */
	if( psent > 0 )
	  avgsent = 1.0*bsent/psent;
	else avgsent = 0.0;
	printf(" %3.1f", avgsent);
	if( precd > 0 )
	  avgrecd = 1.0*brecd/precd;
	else avgrecd = 0.0;
	printf(" %3.1f", avgrecd);

	/* tag the lines that got their IPs swapped to help with debugging */
	if( swapped )
	  printf(" *");
	printf("\n");

	/* track time changes */
	oldtime = newtime;

	/* clean up vars for the next one, if any */
	bzero(leftip, IP_LEN);
	bzero(rightip, IP_LEN);
	bzero(tempip, IP_LEN);
	bzero(lastip, IP_LEN);
	bzero(protocol, LEN);
	bzero(state, LEN);
      }
  }
