/* Parse a connection log file for Zoom call info */
/* and prepare it for plotting in gnuplot.        */

/* Written by Carey Williamson     September 2021 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Manifest constants */
#define MAX_IP 1000
#define IP_LEN 16
#define LEN 8
#define CLIENT_PREFIX "136.159"   /* UCalgary's /16 prefix */
#define PREFIX_LENGTH 7

/* Verbose debugging when needed */
/* #define DEBUG 1 */

int main()
  {
    double newtime, oldtime, starttime;
    char leftip[IP_LEN], rightip[IP_LEN], tempip[IP_LEN], lastip[IP_LEN];
    char iplist[MAX_IP][IP_LEN];
    int leftport, rightport, tempport;
    int psent, precd, bsent, brecd, temp;
    float duration;
    char protocol[LEN], state[LEN];
    int line, count, swapped, i;
    float avgsent, avgrecd, index;
    float sentrate, recdrate;
    
    /* initialization */
    bzero(leftip, IP_LEN);
    bzero(rightip, IP_LEN);
    bzero(tempip, IP_LEN);
    bzero(lastip, IP_LEN);
    bzero(protocol, LEN);
    bzero(state, LEN);
    line = 0;
    count = 0;
    oldtime = 0.0;

    /* scan the log file */
    while( scanf("%lf %s %d %s %d %s %f %s %d %d %d %d\n",
		 &newtime, leftip, &leftport, rightip, &rightport, protocol,
		 &duration, state, &psent, &bsent, &precd, &brecd) != EOF )
      {
	line++;
	if( line == 1 )
	  starttime = newtime;
#ifdef DEBUG
	printf("protocol field on line %d for host %s to host %s is initially '%s'\n",
	       line, leftip, rightip, protocol);
#endif

	/* make sure that client IP is always on the left (originator) */
	swapped = 0;
	if( (strcmp(leftip, CLIENT_PREFIX) < PREFIX_LENGTH) )
	  {
	    /* switch left IP and right IP */
	    strcpy(tempip, leftip);
	    strcpy(leftip, rightip);
	    strcpy(rightip, tempip);
	    /* port numbers */
	    temp = leftport;
	    leftport = rightport;
	    rightport = temp;
	    /* sent and received values for packets and for bytes */
	    temp = psent;
	    psent = precd;
	    precd = temp;
	    temp = bsent;
	    bsent = brecd;
	    brecd = temp;
	    swapped = 1;
	  }

	/* update list of distinct client IPs seen, if needed */
	index = 0.0;
	for( i = 0; i < count; i++ )
	  {
	    if( strcmp(leftip, iplist[i]) == 0 )
	      {
#ifdef DEBUG
		printf("IP %s is already on iplist in position %d\n", leftip, i);
#endif
		index = i+1.0;  /* for vertical axis in graph */
		break;
	      }
	  }
	if( i == count )
	  {
#ifdef DEBUG
	    printf("adding IP %s to iplist in position %d\n", leftip, i);
#endif
	    strcpy(iplist[i], leftip);
	    count++;
	    index = count+0.0;  /* for vertical axis in graph */
	    if( count == MAX_IP )
	      {
		printf("Warning! MAX_IP %d reached on line %d\n", MAX_IP, line);
		exit(1);
	      }
	  }
#ifdef DEBUG
	printf("protocol field on line %d for host %s to host %s is still '%s'\n",
	       line, leftip, rightip, protocol);
#endif
	/* nudge the index up or down a small displacement to improve graph readability */
	if( strcmp(protocol, "tcp") == 0 )
	  index += 0.1;
	else if( strcmp(protocol, "udp") == 0 )
	  index -= 0.1;
	else index -= 0.0;

	/* print out this line of the log, with relative timestamps in seconds */
	printf("%lf %d %3.1f %s %d %s %d %s %f %f %f %s %d %d %d %d",
	       newtime - starttime, line, index,
	       leftip, leftport, rightip, rightport, protocol,
	       duration, newtime - starttime,
	       newtime - starttime + duration,
	       state, psent, bsent, precd, brecd);

	/* report average packet sizes in each direction */
	if( psent > 0 )
	  avgsent = 1.0*bsent/psent;
	else avgsent = 0.0;
	printf(" %3.1f", avgsent);
	if( precd > 0 )
	  avgrecd = 1.0*brecd/precd;
	else avgrecd = 0.0;
	printf(" %3.1f", avgrecd);

	/* report average packet rates in each direction */
	if( duration > 1.0 )
	  sentrate = 1.0*psent/duration;
	else sentrate = 0.0;
	printf(" %3.1f", sentrate);
	if( duration > 1.0 )
	  recdrate = 1.0*precd/duration;
	else recdrate = 0.0;
	printf(" %3.1f", recdrate);

	/* tag the lines that got their IPs swapped to help with debugging */
	if( swapped )
	  printf(" *");
	printf("\n");

	/* track time changes */
	oldtime = newtime;

	/* clean up vars for the next one, if any */
	bzero(leftip, IP_LEN);
	bzero(rightip, IP_LEN);
	bzero(tempip, IP_LEN);
	bzero(lastip, IP_LEN);
	bzero(protocol, LEN);
	bzero(state, LEN);
      }
  }
